#!/usr/bin/env python

# Copyright 2022 daohu527 <daohu527@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from google.protobuf import descriptor_pb2

from cyber_record.cyber.proto import record_pb2, proto_desc_pb2
from cyber_record.common import (
  Section,
  HEADER_LENGTH,
  Compression,
)
from cyber_record.file_object.chunk import Chunk


def to_pb_compress(compression):
  if compression == Compression.NONE:
    return record_pb2.COMPRESS_NONE
  elif compression == Compression.BZ2:
    return record_pb2.COMPRESS_BZ2
  elif compression == Compression.LZ4:
    return record_pb2.COMPRESS_LZ4
  else:
    raise Exception()


def get_proto_desc(file_descriptor, proto_desc):
  file_desc_proto = descriptor_pb2.FileDescriptorProto()
  file_descriptor.CopyToProto(file_desc_proto)
  proto_desc.desc = file_desc_proto.SerializeToString()
  # proto_desc.desc = file_descriptor.serialized_pb
  for file_desc in file_descriptor.dependencies:
    depend = proto_desc.dependencies.add()
    get_proto_desc(file_desc, depend)


class Writer():
  def __init__(self, bag) -> None:
    self.bag = bag

    # private
    self._header = record_pb2.Header()
    self._index = record_pb2.Index()
    self._chunk = Chunk()

    self._channel_indexs = {}
    self._chunk_header_indexs = {}
    self._chunk_body_indexs = {}

    self.build_header()

  def build_header(self):
    self._header.major_version = self.bag._major_version
    self._header.minor_version = self.bag._minor_version
    self._header.compress = to_pb_compress(self.bag._compression)
    self._header.chunk_interval = self.bag._chunk_interval
    self._header.segment_interval = self.bag._segment_interval
    self._header.chunk_raw_size = self.bag._chunk_raw_size
    self._header.segment_raw_size = self.bag._segment_raw_size

  def write(self, topic, msg, t, proto_descriptor=None):
    if self._header.begin_time == 0:
      self._header.begin_time = t
    self._header.end_time = t

    self._header.message_number += 1

    if self._need_split_chunk():
      self.flush()
      self._chunk.clear()
      self._header.chunk_number += 1

    if self._new_channel(topic):
      proto_desc = proto_desc_pb2.ProtoDesc()
      get_proto_desc(msg.DESCRIPTOR.file, proto_desc)
      if proto_descriptor is None:
        proto_descriptor = proto_desc

      channel_position = self._cur_position()
      self._add_channel(topic, msg, proto_descriptor)
      self._add_channel_index(channel_position, topic, msg, proto_descriptor)
      self._header.channel_number += 1

    # channel_cache.message_number
    self._channel_indexs[topic].message_number += 1

    self._chunk.add_message(topic, msg, t)

  def write_header(self):
    self.write_proto_record(self._header)

  def write_proto_record(self, proto_record):
    if isinstance(proto_record, record_pb2.Header):
      section_type = record_pb2.SECTION_HEADER
      self._set_position(0)
    elif isinstance(proto_record, record_pb2.Index):
      section_type = record_pb2.SECTION_INDEX
      self._header.index_position = self._cur_position()
    elif isinstance(proto_record, record_pb2.ChunkHeader):
      section_type = record_pb2.SECTION_CHUNK_HEADER
    elif isinstance(proto_record, record_pb2.ChunkBody):
      section_type = record_pb2.SECTION_CHUNK_BODY
    elif isinstance(proto_record, record_pb2.Channel):
      section_type = record_pb2.SECTION_CHANNEL
    else:
      raise Exception()

    record = proto_record.SerializeToString()
    section = Section(section_type, len(record))
    self._write_section(section)
    self._write(record)

    if section_type == record_pb2.SECTION_HEADER:
      if HEADER_LENGTH < len(record):
        raise Exception()
      reserved_bytes = bytes(HEADER_LENGTH - len(record))
      self._write(reserved_bytes)

    # Get the position of the end of the file
    self._header.size = self._cur_position()


  def flush(self):
    if not self._chunk.empty():
      chunk_header_position = self._cur_position()
      self.write_proto_record(self._chunk._proto_chunk_header)
      self._add_chunk_header_index(chunk_header_position,
                                   self._chunk._proto_chunk_header)

      chunk_body_position = self._cur_position()
      self.write_proto_record(self._chunk._proto_chunk_body)
      self._add_chunk_body_index(chunk_body_position, self._chunk.num())

      self._header.chunk_number += 1

  def close(self):
    self.flush()
    self.write_proto_record(self._index)

    self._header.is_complete = True
    self.write_header()

  def _add_channel(self, topic, msg, proto_desc):
    proto_channel = record_pb2.Channel()
    proto_channel.name = topic
    proto_channel.message_type = '{}.{}'.format(msg.DESCRIPTOR.file.package, type(msg).__name__)
    proto_channel.proto_desc = proto_desc.SerializeToString()
    self.write_proto_record(proto_channel)

  def _add_channel_index(self, channel_position, topic, msg, proto_desc):
    channel_index = self._index.indexes.add()
    channel_index.type = record_pb2.SECTION_CHANNEL
    channel_index.position = channel_position
    channel_index.channel_cache.message_number = 0
    channel_index.channel_cache.name = topic
    channel_index.channel_cache.message_type = '{}.{}'.format(msg.DESCRIPTOR.file.package, type(msg).__name__)
    channel_index.channel_cache.proto_desc = proto_desc.SerializeToString()

    # add to dict
    self._channel_indexs[topic] = channel_index.channel_cache

  def _add_chunk_header_index(self, chunk_header_position, proto_chunk_header):
    chunk_header_index = self._index.indexes.add()
    chunk_header_index.type = record_pb2.SECTION_CHUNK_HEADER
    chunk_header_index.position = chunk_header_position

    chunk_header_cache = chunk_header_index.chunk_header_cache
    chunk_header_cache.message_number = proto_chunk_header.message_number
    chunk_header_cache.begin_time = proto_chunk_header.begin_time
    chunk_header_cache.end_time = proto_chunk_header.end_time
    chunk_header_cache.raw_size = proto_chunk_header.raw_size

  def _add_chunk_body_index(self, chunk_body_position, message_number):
    chunk_body_index = self._index.indexes.add()
    chunk_body_index.type = record_pb2.SECTION_CHUNK_BODY
    chunk_body_index.position = chunk_body_position
    chunk_body_index.chunk_body_cache.message_number = message_number

  def _need_split_chunk(self):
    return (self._chunk.size() >= self._header.chunk_raw_size or
            self._chunk.interval() >= self._header.chunk_interval)

  def _need_split_file(self):
    return ((self._header.end_time - self._header.begin_time >=
            self._header.segment_interval) or
            (self._header.size >= self._header.segment_raw_size))

  def _new_channel(self, topic):
    return topic not in self._channel_indexs

  def _write_section(self, section):
    section_type = (section.type).to_bytes(8, byteorder='little')
    section_size = (section.size).to_bytes(8, byteorder='little')
    self._write(section_type + section_size)

  def _write(self, data):
    self.bag._file.write(data)

  def _set_position(self, position):
    self.bag._file.seek(position)

  def _cur_position(self):
    return self.bag._file.tell()

  def _skip_size(self, data_size):
    self.bag._file.seek(data_size, 1)



#!/usr/bin/env python

# Copyright 2022 daohu527 <daohu527@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from cyber_record.record import Record


file_name = "example.record.00000"

def read_all():
  record = Record(file_name)
  for topic, message, t in record.read_messages_fallback():
    print("{}, {}, {}".format(topic, type(message), t))


def read_filter_by_topic():
  record = Record(file_name)
  for topic, message, t in record.read_messages('/apollo/sensor/rs32/Scan'):
    print("{}, {}, {}".format(topic, type(message), t))


def read_filter_by_time():
  record = Record(file_name)
  for topic, message, t in record.read_messages(start_time=1627031535164278940,\
      end_time=1627031535215164773):
    print("{}, {}, {}".format(topic, type(message), t))


def read_filter_by_both():
  record = Record(file_name)
  for topic, message, t in record.read_messages('/apollo/canbus/chassis', \
      start_time=1627031535164278940, end_time=1627031535215164773):
    print("{}, {}, {}".format(topic, type(message), t))


if __name__ == "__main__":
  read_all()
  read_filter_by_topic()
  read_filter_by_time()
  read_filter_by_both()



#!/usr/bin/env python

# Copyright 2022 daohu527 <daohu527@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from enum import Enum

SECTION_LENGTH = 16
HEADER_LENGTH = 2048
CHUNK_INTERVAL = 20 * 1000 * 1000 * 1000    # 20s
SEGMENT_INTERVAL = 60 * 1000 * 1000 * 1000  # 60s
CHUNK_RAW_SIZE = 200 * 1024 * 1024     # 200MB
SEGMENT_RAW_SIZE = 2048 * 1024 * 1024  # 2GB

MIN_CHUNK_SIZE = 512

RECORD_MAJOR_VERSION = 1
RECORD_MINOR_VERSION = 0


class Compression(Enum):
  NONE = 'none'
  BZ2  = 'bz2'
  LZ4 = 'lz4'

class Section:
  def __init__(self, section_type=None, data_size=0) -> None:
    self.type = section_type
    self.size = data_size

  def __str__(self):
    return "Section type: {}, size: {}".format(self.type, self.size)



#!/usr/bin/env python

# Copyright 2022 daohu527 <daohu527@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from google.protobuf import message_factory, descriptor_pb2, descriptor_pool


from cyber_record.cyber.proto import record_pb2, proto_desc_pb2
from cyber_record.file_object.chunk import Chunk
from cyber_record.record_exception import RecordException
from cyber_record.common import Section, SECTION_LENGTH, HEADER_LENGTH


class Reader:
  def __init__(self, bag) -> None:
    self.bag = bag
    self.chunk_header_indexs = []
    self.chunk_body_indexs = []
    self.sorted_chunk_indexs = []
    self.channels = {}
    self.desc_pool = descriptor_pool.DescriptorPool()

    self.message_type_pool = {}
    self.chunk = Chunk()
    self.message_index = 0

  def _fill_header(self, header):
    self.bag._version = "{}.{}".format(header.major_version, \
        header.minor_version)
    self.bag._size = header.size
    self.bag._message_number = header.message_number
    self.bag._start_time = header.begin_time
    self.bag._end_time = header.end_time

  def _sort_chunk_indexs(self):
    self.sorted_chunk_indexs = sorted( \
      zip(self.chunk_header_indexs, self.chunk_body_indexs), \
      key=lambda x: x[0].chunk_header_cache.begin_time)

  def start_reading(self):
    header = self.read_header()
    self._fill_header(header)
    print(header)

    index = self.read_index(header)
    for single_index in index.indexes:
      if single_index.type == record_pb2.SECTION_CHUNK_HEADER:
        self.chunk_header_indexs.append(single_index)
      elif single_index.type == record_pb2.SECTION_CHUNK_BODY:
        self.chunk_body_indexs.append(single_index)
      elif single_index.type == record_pb2.SECTION_CHANNEL:
        name = single_index.channel_cache.name
        self.channels[name] = single_index.channel_cache
      else:
        print("Unknown Index type!")

    self._sort_chunk_indexs()
    # print(index)

    self._create_message_type_pool()

    self._set_position(HEADER_LENGTH + SECTION_LENGTH)

  def reindex(self):
    pass

  def get_channel_cache(self, topic_filters):
    filtered_channel_cache = []
    for channel_name in self.channels:
      if topic_filters is None or channel_name not in topic_filters:
        filtered_channel_cache.append(self.channels[channel_name])
    return filtered_channel_cache

  def _is_valid_topic(self, topic, topics):
    if topics is None:
      return True

    return topic in set(topics)

  def _is_valid_time(self, cur_time, start_time, end_time):
    if start_time and cur_time < start_time:
      return False
    if end_time and cur_time > end_time:
      return False
    return True

  def _get_chunk_body_indexs(self, start_time, end_time):
    for chunk_header_index, chunk_body_index in self.sorted_chunk_indexs:
      if start_time and \
          chunk_header_index.chunk_header_cache.end_time < start_time:
        continue

      if end_time and \
          chunk_header_index.chunk_header_cache.begin_time > end_time:
        continue
      # Todo(zero): should be chunk_body_index, there maybe a bug in apollo!!!
      yield chunk_body_index

  def read_messages(self, topics, start_time, end_time):
    for chunk_body_index in self._get_chunk_body_indexs(start_time, end_time):
      # print(chunk_body_index)
      proto_chunk_body = self.read_chunk_body(chunk_body_index.position)
      if proto_chunk_body is None:
        continue
      self.chunk.swap(proto_chunk_body)

      while not self.chunk.end():
        single_message = self.chunk.next_message()
        if self._is_valid_topic(single_message.channel_name, topics) and \
           self._is_valid_time(single_message.time, start_time, end_time):
          proto_message = self._create_message(single_message)
          yield single_message.channel_name, proto_message, single_message.time

  def read_messages_fallback(self, topics, start_time, end_time):
    """
    deprecated
    """
    while self.message_index < self.bag._message_number:
      if self.chunk.end():
        self._read_next_chunk()

      single_message = self.chunk.next_message()
      proto_message = self._create_message(single_message)
      self.message_index += 1
      yield single_message.channel_name, proto_message, single_message.time

  def read_header(self):
    self._set_position(0)

    section = Section()
    self._read_section(section)

    if section.type != record_pb2.SECTION_HEADER:
      return None

    proto_header = record_pb2.Header()
    data = self._read(section.size)

    proto_header.ParseFromString(data)

    self._set_position(HEADER_LENGTH + SECTION_LENGTH)
    return proto_header

  def read_index(self, header):
    self._set_position(header.index_position)

    section = Section()
    self._read_section(section)

    if section.type != record_pb2.SECTION_INDEX:
      return None

    proto_index = record_pb2.Index()
    data = self._read(section.size)

    proto_index.ParseFromString(data)
    return proto_index

  def read_chunk_header(self, position):
    self._set_position(position)

    section = Section()
    self._read_section(section)

    if section.type != record_pb2.SECTION_CHUNK_HEADER:
      return None

    chunk_header = record_pb2.ChunkHeader()
    data = self._read(section.size)

    chunk_header.ParseFromString(data)
    return chunk_header

  def read_chunk_body(self, position):
    self._set_position(position)

    section = Section()
    self._read_section(section)

    if section.type != record_pb2.SECTION_CHUNK_BODY:
      return None

    chunk_body = record_pb2.ChunkBody()
    data = self._read(section.size)

    chunk_body.ParseFromString(data)
    return chunk_body

  def _read_section(self, section):
    section.type = int.from_bytes(self._read(4), byteorder='little')
    self._skip_size(4)
    section.size = int.from_bytes(self._read(8), byteorder='little')
    # print(section)

  def _read_next_chunk(self):
    while self.bag._file.tell() != self.bag._size:
      section = Section()
      self._read_section(section)

      if section.type == record_pb2.SECTION_CHUNK_BODY:
        data = self._read(section.size)
        proto_chunk_body = record_pb2.ChunkBody()
        proto_chunk_body.ParseFromString(data)
        self.chunk.swap(proto_chunk_body)
        return True
      else:
        self._skip_size(section.size)
    else:
      return False

  def _add_dependency(self, proto_desc):
    if proto_desc is None:
      return

    file_desc_proto = descriptor_pb2.FileDescriptorProto()
    file_desc_proto.ParseFromString(proto_desc.desc)
    for dependency in proto_desc.dependencies:
      self._add_dependency(dependency)
    self.desc_pool.Add(file_desc_proto)
    # print(file_desc_proto)

  def _create_message_type_pool(self):
    for channel_name, channel_cache in self.channels.items():
      proto_desc = proto_desc_pb2.ProtoDesc()
      proto_desc.ParseFromString(channel_cache.proto_desc)
      self._add_dependency(proto_desc)

      # print(channel_cache.message_type)
      descriptor = self.desc_pool.FindMessageTypeByName(channel_cache.message_type)
      message_type = message_factory.MessageFactory().GetPrototype(descriptor)
      self.message_type_pool.update({channel_name: message_type})

  def _create_message(self, single_message):
    message_type = self.message_type_pool.get(single_message.channel_name, None)

    if message_type is None:
      return None
    proto_message = message_type()
    proto_message.ParseFromString(single_message.content)

    return proto_message

  def _read(self, size):
    data = self.bag._file.read(size)
    if len(data) != size:
      raise RecordException('expecting {} bytes, read {}'.format(size, len(data)))
    return data

  def _set_position(self, position):
    self.bag._file.seek(position)

  def _cur_position(self):
    return self.bag._file.tell()

  def _skip_size(self, data_size):
    self.bag._file.seek(data_size, 1)

  def _skip_record(self):
    section = Section()
    self._read_section(section)
    self._skip_size(section.size)



#!/usr/bin/env python

# Copyright 2022 daohu527 <daohu527@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from cyber_record.cyber.proto import record_pb2


class Chunk:
  def __init__(self) -> None:
    self._index = 0
    self._proto_chunk_header = record_pb2.ChunkHeader()
    self._proto_chunk_body = record_pb2.ChunkBody()

  def swap(self, proto_chunk_body):
    self._index = 0
    self._proto_chunk_body = proto_chunk_body
    self._proto_chunk_header.message_number = len(proto_chunk_body.messages)

  def clear(self):
    self._index = 0
    self._proto_chunk_header = record_pb2.ChunkHeader()
    self._proto_chunk_body = record_pb2.ChunkBody()

  def next_message(self):
    if self._index >= self.num() or self._index < 0:
      return None

    message = self._proto_chunk_body.messages[self._index]
    self._index += 1
    return message

  def end(self):
    return self._index >= self.num()

  def add_message(self, topic, msg, t, raw=True):
    if self._proto_chunk_header.begin_time == 0:
      self._proto_chunk_header.begin_time = t
    self._proto_chunk_header.end_time = t

    message = self._proto_chunk_body.messages.add()
    message.channel_name = topic
    message.time = t
    message.content = msg.SerializeToString()

    self._proto_chunk_header.raw_size += len(message.content)
    self._proto_chunk_header.message_number += 1

  def empty(self):
    return self.num() == 0

  def interval(self):
    return (self._proto_chunk_header.end_time -
            self._proto_chunk_header.begin_time)

  def size(self):
    return self._proto_chunk_header.raw_size

  def num(self):
    return self._proto_chunk_header.message_number



#!/usr/bin/env python

# Copyright 2022 daohu527 <daohu527@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import csv

from cyber_record.record import Record
from record_msg.parser import to_csv, ImageParser, PointCloudParser


def parse_pose(pose):
  '''
  save pose to csv file
  '''
  line = to_csv([pose.header.timestamp_sec, pose.pose])
  writer.writerow(line)

def parse_chassis(chassis):
  pass

def parse_prediction(prediction):
  pass

def parse_routing_response_history(routing_response_history):
  pass

def parse_planning(planning):
  pass

def parse_image(image):
  image_parser.parse(image)

def parse_pointcloud(pointcloud):
  pointcloud_parser.parse(message)

if __name__ == "__main__":
  # csv
  f = open("../tests/message.csv", 'w')
  writer = csv.writer(f)

  # bag
  file_name = "example.record.00000"
  record = Record(file_name)

  image_parser = ImageParser('../tests')
  pointcloud_parser = PointCloudParser('../tests')

  for topic, message, t in record.read_messages_fallback():
    if topic == "/apollo/localization/pose":
      parse_pose(message)
    elif topic == "/apollo/canbus/chassis":
      parse_chassis(message)
    elif topic == "/apollo/prediction":
      parse_prediction(message)
    elif topic == "/apollo/routing_response_history":
      parse_routing_response_history(message)
    elif topic == "/apollo/planning":
      parse_planning(message)
    elif topic == "/apollo/sensor/camera/front_6mm/image":
      parse_image(message)
    elif topic == "/apollo/sensor/lidar32/compensator/PointCloud2":
      parse_pointcloud(message)
    else:
      pass

  f.close()



#!/usr/bin/env python

# Copyright 2022 daohu527 <daohu527@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import struct
import io
import os.path
import time


from cyber_record.reader import Reader
from cyber_record.writer import Writer

from cyber_record.common import (
  Compression,
  CHUNK_RAW_SIZE,
  SEGMENT_RAW_SIZE,
  CHUNK_INTERVAL,
  SEGMENT_INTERVAL,
  MIN_CHUNK_SIZE,
  RECORD_MAJOR_VERSION,
  RECORD_MINOR_VERSION,
)


class Record(object):
  """
  Serialize messages to and from a single file on disk using record format.
  """
  def __init__(self, f, mode='r', compression=Compression.NONE, \
      chunk_threshold=CHUNK_RAW_SIZE, allow_unindexed=False, \
      options=None):
    """
    Open a bag file.  The mode can be 'r', 'w', or 'a' for reading (default),
    writing or appending.  The file will be created if it doesn't exist
    when opened for writing or appending; it will be truncated when opened
    for writing.  Simultaneous reading and writing is allowed when in writing
    or appending mode.

    Args:
        f (_type_): _description_
        mode (str, optional): _description_. Defaults to 'r'.
        compression (_type_, optional): _description_. Defaults to Compression.NONE.
        allow_unindexed (bool, optional): _description_. Defaults to False.

    Raises:
        ValueError: _description_
    """
    # options
    if options is not None:
      if not isinstance(options, dict):
        raise ValueError('options must be of type dict')
      if 'compression' in options:
        compression = options['compression']
      if 'chunk_threshold' in options:
        chunk_threshold = options['chunk_threshold']

    self._file           = None
    self._filename       = None
    self._major_version  = RECORD_MAJOR_VERSION
    self._minor_version  = RECORD_MINOR_VERSION

    self._size           = 0
    self._message_number = 0

    self._start_time = 0
    self._end_time   = 0


    assert chunk_threshold >= MIN_CHUNK_SIZE, \
        "chunk_threshold should large than {}".format(MIN_CHUNK_SIZE)
    self._chunk_threshold = chunk_threshold

    # config
    self._chunk_interval = CHUNK_INTERVAL
    self._chunk_raw_size = chunk_threshold
    self._segment_interval = SEGMENT_INTERVAL
    self._segment_raw_size = SEGMENT_RAW_SIZE

    allowed_compressions = set(item for item in Compression)
    if compression not in allowed_compressions:
      raise ValueError('compression must be one of: {}'.format(allowed_compressions))
    self._compression = compression

    self._reader = None
    self._writer = None
    self._encryptor = None

    self._open(f, mode, allow_unindexed)


  def __iter__(self):
    return self.read_messages()

  def __enter__(self):
    return self

  def __exit__(self, exc_type, exc_value, traceback):
    self.close()

  @property
  def options(self):
    return {'compression' : self._compression, \
            'chunk_threshold' : self._chunk_threshold}

  @property
  def filename(self):
    return self._filename

  @property
  def version(self):
    return "{}.{}".format(self._major_version, self._minor_version)

  @property
  def mode(self):
    return self._mode

  @property
  def size(self):
    return self._size

  def _get_compression(self):
    return self._compression

  def _set_compression(self, compression):
    allowed_compressions = set(item.value for item in Compression)
    if compression not in allowed_compressions:
      raise ValueError('compression must be one of: {}'.format(allowed_compressions))

    self._compression = compression

  compression = property(_get_compression, _set_compression)


  def _get_chunk_threshold(self):
    return self._chunk_threshold

  def _set_chunk_threshold(self, chunk_threshold):
    assert chunk_threshold >= MIN_CHUNK_SIZE, \
        "chunk_threshold should large than {}".format(MIN_CHUNK_SIZE)
    self._chunk_threshold = chunk_threshold

  chunk_threshold = property(_get_chunk_threshold, _set_chunk_threshold)


  def read_messages(self, topics=None, start_time=None, end_time=None):
    if topics and isinstance(topics, str):
      topics = [topics]

    return self._reader.read_messages(topics, start_time, end_time)

  def read_messages_fallback(self, topics=None, start_time=None, end_time=None):
    """
    deprecated
    """
    if topics and isinstance(topics, str):
      topics = [topics]

    return self._reader.read_messages_fallback(topics, start_time, end_time)

  def write(self, topic, msg, t=None, proto_descriptor=None):
    if not self._file:
      raise ValueError('I/O operation on closed record')

    if not topic:
      raise ValueError('topic is invalid')
    if not msg:
      raise ValueError('msg is invalid')

    if t is None:
      time_ns = getattr(time, "time_ns", None)
      if callable(time_ns):
        t = time.time_ns()
      else:
        t = int(time.time() * 1e9)

    if self._writer._need_split_file():
      # Todo(zero): need replace file handle, we don't support yet!
      pass

    self._writer.write(topic, msg, t, proto_descriptor)

  def reindex(self):
    # todo(zero): Reindex, modify chunkinfo and descriptor
    pass

  def close(self):
    if self._file:
      if self._mode in 'wa':
        self._stop_writing()

      self._close_file()

  def get_message_count(self, topic_filters=None):
    message_count = 0

    if topic_filters is not None:
      channel_cache = self._reader.get_channel_cache(topic_filters)
      for channel_info in channel_cache:
        message_count += channel_info.message_number
    else:
      message_count = self._message_number

    return message_count

  def get_start_time(self):
    return self._start_time

  def get_end_time(self):
    return self._end_time

  def set_encryptor(self, encryptor=None, param=None):
    pass

  def __str__(self):
    pass


  # internal interface
  def _get_descriptors(self, topics=None, descriptor_filter=None):
    pass

  def _get_entries(self, descriptors=None, start_time=None, end_time=None):
    pass

  def _get_entry(self, t, descriptors=None):
    pass

  def _clear_index(self):
    pass

  def _open(self, f, mode, allow_unindexed):
    assert f is not None, "filename (or stream) is invalid"

    self._mode = mode

    try:
      if mode == 'r': self._open_read(f)
      elif mode == 'w': self._open_write(f)
      elif mode == 'a': self._open_append(f)
      else:
        raise ValueError('mode {} is invalid'.format(mode))
    except struct.error:
      raise Exception()

  def _is_file(self, f):
    return isinstance(f, io.IOBase)

  def _open_read(self, f):
    if self._is_file(f):
      self._file = f
      self._filename = None
    else:
      self._file = open(f, 'rb')
      self._filename = f

    try:
      self._create_reader()
      self._reader.start_reading()
    except:
      self._close_file()
      raise

  def _open_write(self, f):
    if self._is_file(f):
      self._file = f
      self._filename = None
    else:
      self._file = open(f, 'w+b')
      self._filename = f

    try:
      self._create_writer()
      self._start_writing()
    except:
      self._close_file()
      raise

  def _open_append(self, f):
    if self._is_file(f):
      self._file = f
      self._filename = None
    else:
      self._filename = f
      # if file exists
      if os.path.isfile(f):
        self._file = open(f, 'r+b')
      else:
        self._file = open(f, 'w+b')

    try:
      self._start_appending()
    except:
      self._close_file()
      raise

  def _close_file(self):
    self._file.close()
    self._file = None

  def _create_reader(self):
    self._reader = Reader(self)

  def _create_writer(self):
    self._writer = Writer(self)

  def _start_writing(self):
    self._writer.write_header()

  def _start_appending(self):
    pass

  def _stop_writing(self):
    self._writer.close()



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: cyber/proto/proto_desc.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='cyber/proto/proto_desc.proto',
  package='apollo.cyber.proto',
  syntax='proto2',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x1c\x63yber/proto/proto_desc.proto\x12\x12\x61pollo.cyber.proto\"N\n\tProtoDesc\x12\x0c\n\x04\x64\x65sc\x18\x01 \x01(\x0c\x12\x33\n\x0c\x64\x65pendencies\x18\x02 \x03(\x0b\x32\x1d.apollo.cyber.proto.ProtoDesc'
)




_PROTODESC = _descriptor.Descriptor(
  name='ProtoDesc',
  full_name='apollo.cyber.proto.ProtoDesc',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='desc', full_name='apollo.cyber.proto.ProtoDesc.desc', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='dependencies', full_name='apollo.cyber.proto.ProtoDesc.dependencies', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=52,
  serialized_end=130,
)

_PROTODESC.fields_by_name['dependencies'].message_type = _PROTODESC
DESCRIPTOR.message_types_by_name['ProtoDesc'] = _PROTODESC
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

ProtoDesc = _reflection.GeneratedProtocolMessageType('ProtoDesc', (_message.Message,), {
  'DESCRIPTOR' : _PROTODESC,
  '__module__' : 'cyber.proto.proto_desc_pb2'
  # @@protoc_insertion_point(class_scope:apollo.cyber.proto.ProtoDesc)
  })
_sym_db.RegisterMessage(ProtoDesc)


# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: cyber/proto/record.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='cyber/proto/record.proto',
  package='apollo.cyber.proto',
  syntax='proto2',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x18\x63yber/proto/record.proto\x12\x12\x61pollo.cyber.proto\"\x96\x02\n\x0bSingleIndex\x12-\n\x04type\x18\x01 \x01(\x0e\x32\x1f.apollo.cyber.proto.SectionType\x12\x10\n\x08position\x18\x02 \x01(\x04\x12\x39\n\rchannel_cache\x18\x65 \x01(\x0b\x32 .apollo.cyber.proto.ChannelCacheH\x00\x12\x42\n\x12\x63hunk_header_cache\x18\x66 \x01(\x0b\x32$.apollo.cyber.proto.ChunkHeaderCacheH\x00\x12>\n\x10\x63hunk_body_cache\x18g \x01(\x0b\x32\".apollo.cyber.proto.ChunkBodyCacheH\x00\x42\x07\n\x05\x63\x61\x63he\"b\n\x10\x43hunkHeaderCache\x12\x16\n\x0emessage_number\x18\x01 \x01(\x04\x12\x12\n\nbegin_time\x18\x02 \x01(\x04\x12\x10\n\x08\x65nd_time\x18\x03 \x01(\x04\x12\x10\n\x08raw_size\x18\x04 \x01(\x04\"(\n\x0e\x43hunkBodyCache\x12\x16\n\x0emessage_number\x18\x01 \x01(\x04\"^\n\x0c\x43hannelCache\x12\x16\n\x0emessage_number\x18\x01 \x01(\x04\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x14\n\x0cmessage_type\x18\x03 \x01(\t\x12\x12\n\nproto_desc\x18\x04 \x01(\x0c\"D\n\rSingleMessage\x12\x14\n\x0c\x63hannel_name\x18\x01 \x01(\t\x12\x0c\n\x04time\x18\x02 \x01(\x04\x12\x0f\n\x07\x63ontent\x18\x03 \x01(\x0c\"\x91\x03\n\x06Header\x12\x15\n\rmajor_version\x18\x01 \x01(\r\x12\x15\n\rminor_version\x18\x02 \x01(\r\x12\x32\n\x08\x63ompress\x18\x03 \x01(\x0e\x32 .apollo.cyber.proto.CompressType\x12\x16\n\x0e\x63hunk_interval\x18\x04 \x01(\x04\x12\x18\n\x10segment_interval\x18\x05 \x01(\x04\x12\x19\n\x0eindex_position\x18\x06 \x01(\x04:\x01\x30\x12\x17\n\x0c\x63hunk_number\x18\x07 \x01(\x04:\x01\x30\x12\x19\n\x0e\x63hannel_number\x18\x08 \x01(\x04:\x01\x30\x12\x15\n\nbegin_time\x18\t \x01(\x04:\x01\x30\x12\x13\n\x08\x65nd_time\x18\n \x01(\x04:\x01\x30\x12\x19\n\x0emessage_number\x18\x0b \x01(\x04:\x01\x30\x12\x0f\n\x04size\x18\x0c \x01(\x04:\x01\x30\x12\x1a\n\x0bis_complete\x18\r \x01(\x08:\x05\x66\x61lse\x12\x16\n\x0e\x63hunk_raw_size\x18\x0e \x01(\x04\x12\x18\n\x10segment_raw_size\x18\x0f \x01(\x04\"A\n\x07\x43hannel\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x14\n\x0cmessage_type\x18\x02 \x01(\t\x12\x12\n\nproto_desc\x18\x03 \x01(\x0c\"]\n\x0b\x43hunkHeader\x12\x12\n\nbegin_time\x18\x01 \x01(\x04\x12\x10\n\x08\x65nd_time\x18\x02 \x01(\x04\x12\x16\n\x0emessage_number\x18\x03 \x01(\x04\x12\x10\n\x08raw_size\x18\x04 \x01(\x04\"@\n\tChunkBody\x12\x33\n\x08messages\x18\x01 \x03(\x0b\x32!.apollo.cyber.proto.SingleMessage\"9\n\x05Index\x12\x30\n\x07indexes\x18\x01 \x03(\x0b\x32\x1f.apollo.cyber.proto.SingleIndex*{\n\x0bSectionType\x12\x12\n\x0eSECTION_HEADER\x10\x00\x12\x18\n\x14SECTION_CHUNK_HEADER\x10\x01\x12\x16\n\x12SECTION_CHUNK_BODY\x10\x02\x12\x11\n\rSECTION_INDEX\x10\x03\x12\x13\n\x0fSECTION_CHANNEL\x10\x04*E\n\x0c\x43ompressType\x12\x11\n\rCOMPRESS_NONE\x10\x00\x12\x10\n\x0c\x43OMPRESS_BZ2\x10\x01\x12\x10\n\x0c\x43OMPRESS_LZ4\x10\x02'
)

_SECTIONTYPE = _descriptor.EnumDescriptor(
  name='SectionType',
  full_name='apollo.cyber.proto.SectionType',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='SECTION_HEADER', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='SECTION_CHUNK_HEADER', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='SECTION_CHUNK_BODY', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='SECTION_INDEX', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='SECTION_CHANNEL', index=4, number=4,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=1328,
  serialized_end=1451,
)
_sym_db.RegisterEnumDescriptor(_SECTIONTYPE)

SectionType = enum_type_wrapper.EnumTypeWrapper(_SECTIONTYPE)
_COMPRESSTYPE = _descriptor.EnumDescriptor(
  name='CompressType',
  full_name='apollo.cyber.proto.CompressType',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='COMPRESS_NONE', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='COMPRESS_BZ2', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='COMPRESS_LZ4', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=1453,
  serialized_end=1522,
)
_sym_db.RegisterEnumDescriptor(_COMPRESSTYPE)

CompressType = enum_type_wrapper.EnumTypeWrapper(_COMPRESSTYPE)
SECTION_HEADER = 0
SECTION_CHUNK_HEADER = 1
SECTION_CHUNK_BODY = 2
SECTION_INDEX = 3
SECTION_CHANNEL = 4
COMPRESS_NONE = 0
COMPRESS_BZ2 = 1
COMPRESS_LZ4 = 2



_SINGLEINDEX = _descriptor.Descriptor(
  name='SingleIndex',
  full_name='apollo.cyber.proto.SingleIndex',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='type', full_name='apollo.cyber.proto.SingleIndex.type', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='position', full_name='apollo.cyber.proto.SingleIndex.position', index=1,
      number=2, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='channel_cache', full_name='apollo.cyber.proto.SingleIndex.channel_cache', index=2,
      number=101, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='chunk_header_cache', full_name='apollo.cyber.proto.SingleIndex.chunk_header_cache', index=3,
      number=102, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='chunk_body_cache', full_name='apollo.cyber.proto.SingleIndex.chunk_body_cache', index=4,
      number=103, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='cache', full_name='apollo.cyber.proto.SingleIndex.cache',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=49,
  serialized_end=327,
)


_CHUNKHEADERCACHE = _descriptor.Descriptor(
  name='ChunkHeaderCache',
  full_name='apollo.cyber.proto.ChunkHeaderCache',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='message_number', full_name='apollo.cyber.proto.ChunkHeaderCache.message_number', index=0,
      number=1, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='begin_time', full_name='apollo.cyber.proto.ChunkHeaderCache.begin_time', index=1,
      number=2, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='end_time', full_name='apollo.cyber.proto.ChunkHeaderCache.end_time', index=2,
      number=3, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='raw_size', full_name='apollo.cyber.proto.ChunkHeaderCache.raw_size', index=3,
      number=4, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=329,
  serialized_end=427,
)


_CHUNKBODYCACHE = _descriptor.Descriptor(
  name='ChunkBodyCache',
  full_name='apollo.cyber.proto.ChunkBodyCache',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='message_number', full_name='apollo.cyber.proto.ChunkBodyCache.message_number', index=0,
      number=1, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=429,
  serialized_end=469,
)


_CHANNELCACHE = _descriptor.Descriptor(
  name='ChannelCache',
  full_name='apollo.cyber.proto.ChannelCache',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='message_number', full_name='apollo.cyber.proto.ChannelCache.message_number', index=0,
      number=1, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='name', full_name='apollo.cyber.proto.ChannelCache.name', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='message_type', full_name='apollo.cyber.proto.ChannelCache.message_type', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='proto_desc', full_name='apollo.cyber.proto.ChannelCache.proto_desc', index=3,
      number=4, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=471,
  serialized_end=565,
)


_SINGLEMESSAGE = _descriptor.Descriptor(
  name='SingleMessage',
  full_name='apollo.cyber.proto.SingleMessage',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='channel_name', full_name='apollo.cyber.proto.SingleMessage.channel_name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='time', full_name='apollo.cyber.proto.SingleMessage.time', index=1,
      number=2, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='content', full_name='apollo.cyber.proto.SingleMessage.content', index=2,
      number=3, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=567,
  serialized_end=635,
)


_HEADER = _descriptor.Descriptor(
  name='Header',
  full_name='apollo.cyber.proto.Header',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='major_version', full_name='apollo.cyber.proto.Header.major_version', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='minor_version', full_name='apollo.cyber.proto.Header.minor_version', index=1,
      number=2, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='compress', full_name='apollo.cyber.proto.Header.compress', index=2,
      number=3, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='chunk_interval', full_name='apollo.cyber.proto.Header.chunk_interval', index=3,
      number=4, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='segment_interval', full_name='apollo.cyber.proto.Header.segment_interval', index=4,
      number=5, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='index_position', full_name='apollo.cyber.proto.Header.index_position', index=5,
      number=6, type=4, cpp_type=4, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='chunk_number', full_name='apollo.cyber.proto.Header.chunk_number', index=6,
      number=7, type=4, cpp_type=4, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='channel_number', full_name='apollo.cyber.proto.Header.channel_number', index=7,
      number=8, type=4, cpp_type=4, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='begin_time', full_name='apollo.cyber.proto.Header.begin_time', index=8,
      number=9, type=4, cpp_type=4, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='end_time', full_name='apollo.cyber.proto.Header.end_time', index=9,
      number=10, type=4, cpp_type=4, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='message_number', full_name='apollo.cyber.proto.Header.message_number', index=10,
      number=11, type=4, cpp_type=4, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='size', full_name='apollo.cyber.proto.Header.size', index=11,
      number=12, type=4, cpp_type=4, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='is_complete', full_name='apollo.cyber.proto.Header.is_complete', index=12,
      number=13, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='chunk_raw_size', full_name='apollo.cyber.proto.Header.chunk_raw_size', index=13,
      number=14, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='segment_raw_size', full_name='apollo.cyber.proto.Header.segment_raw_size', index=14,
      number=15, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=638,
  serialized_end=1039,
)


_CHANNEL = _descriptor.Descriptor(
  name='Channel',
  full_name='apollo.cyber.proto.Channel',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='apollo.cyber.proto.Channel.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='message_type', full_name='apollo.cyber.proto.Channel.message_type', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='proto_desc', full_name='apollo.cyber.proto.Channel.proto_desc', index=2,
      number=3, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1041,
  serialized_end=1106,
)


_CHUNKHEADER = _descriptor.Descriptor(
  name='ChunkHeader',
  full_name='apollo.cyber.proto.ChunkHeader',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='begin_time', full_name='apollo.cyber.proto.ChunkHeader.begin_time', index=0,
      number=1, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='end_time', full_name='apollo.cyber.proto.ChunkHeader.end_time', index=1,
      number=2, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='message_number', full_name='apollo.cyber.proto.ChunkHeader.message_number', index=2,
      number=3, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='raw_size', full_name='apollo.cyber.proto.ChunkHeader.raw_size', index=3,
      number=4, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1108,
  serialized_end=1201,
)


_CHUNKBODY = _descriptor.Descriptor(
  name='ChunkBody',
  full_name='apollo.cyber.proto.ChunkBody',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='messages', full_name='apollo.cyber.proto.ChunkBody.messages', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1203,
  serialized_end=1267,
)


_INDEX = _descriptor.Descriptor(
  name='Index',
  full_name='apollo.cyber.proto.Index',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='indexes', full_name='apollo.cyber.proto.Index.indexes', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1269,
  serialized_end=1326,
)

_SINGLEINDEX.fields_by_name['type'].enum_type = _SECTIONTYPE
_SINGLEINDEX.fields_by_name['channel_cache'].message_type = _CHANNELCACHE
_SINGLEINDEX.fields_by_name['chunk_header_cache'].message_type = _CHUNKHEADERCACHE
_SINGLEINDEX.fields_by_name['chunk_body_cache'].message_type = _CHUNKBODYCACHE
_SINGLEINDEX.oneofs_by_name['cache'].fields.append(
  _SINGLEINDEX.fields_by_name['channel_cache'])
_SINGLEINDEX.fields_by_name['channel_cache'].containing_oneof = _SINGLEINDEX.oneofs_by_name['cache']
_SINGLEINDEX.oneofs_by_name['cache'].fields.append(
  _SINGLEINDEX.fields_by_name['chunk_header_cache'])
_SINGLEINDEX.fields_by_name['chunk_header_cache'].containing_oneof = _SINGLEINDEX.oneofs_by_name['cache']
_SINGLEINDEX.oneofs_by_name['cache'].fields.append(
  _SINGLEINDEX.fields_by_name['chunk_body_cache'])
_SINGLEINDEX.fields_by_name['chunk_body_cache'].containing_oneof = _SINGLEINDEX.oneofs_by_name['cache']
_HEADER.fields_by_name['compress'].enum_type = _COMPRESSTYPE
_CHUNKBODY.fields_by_name['messages'].message_type = _SINGLEMESSAGE
_INDEX.fields_by_name['indexes'].message_type = _SINGLEINDEX
DESCRIPTOR.message_types_by_name['SingleIndex'] = _SINGLEINDEX
DESCRIPTOR.message_types_by_name['ChunkHeaderCache'] = _CHUNKHEADERCACHE
DESCRIPTOR.message_types_by_name['ChunkBodyCache'] = _CHUNKBODYCACHE
DESCRIPTOR.message_types_by_name['ChannelCache'] = _CHANNELCACHE
DESCRIPTOR.message_types_by_name['SingleMessage'] = _SINGLEMESSAGE
DESCRIPTOR.message_types_by_name['Header'] = _HEADER
DESCRIPTOR.message_types_by_name['Channel'] = _CHANNEL
DESCRIPTOR.message_types_by_name['ChunkHeader'] = _CHUNKHEADER
DESCRIPTOR.message_types_by_name['ChunkBody'] = _CHUNKBODY
DESCRIPTOR.message_types_by_name['Index'] = _INDEX
DESCRIPTOR.enum_types_by_name['SectionType'] = _SECTIONTYPE
DESCRIPTOR.enum_types_by_name['CompressType'] = _COMPRESSTYPE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

SingleIndex = _reflection.GeneratedProtocolMessageType('SingleIndex', (_message.Message,), {
  'DESCRIPTOR' : _SINGLEINDEX,
  '__module__' : 'cyber.proto.record_pb2'
  # @@protoc_insertion_point(class_scope:apollo.cyber.proto.SingleIndex)
  })
_sym_db.RegisterMessage(SingleIndex)

ChunkHeaderCache = _reflection.GeneratedProtocolMessageType('ChunkHeaderCache', (_message.Message,), {
  'DESCRIPTOR' : _CHUNKHEADERCACHE,
  '__module__' : 'cyber.proto.record_pb2'
  # @@protoc_insertion_point(class_scope:apollo.cyber.proto.ChunkHeaderCache)
  })
_sym_db.RegisterMessage(ChunkHeaderCache)

ChunkBodyCache = _reflection.GeneratedProtocolMessageType('ChunkBodyCache', (_message.Message,), {
  'DESCRIPTOR' : _CHUNKBODYCACHE,
  '__module__' : 'cyber.proto.record_pb2'
  # @@protoc_insertion_point(class_scope:apollo.cyber.proto.ChunkBodyCache)
  })
_sym_db.RegisterMessage(ChunkBodyCache)

ChannelCache = _reflection.GeneratedProtocolMessageType('ChannelCache', (_message.Message,), {
  'DESCRIPTOR' : _CHANNELCACHE,
  '__module__' : 'cyber.proto.record_pb2'
  # @@protoc_insertion_point(class_scope:apollo.cyber.proto.ChannelCache)
  })
_sym_db.RegisterMessage(ChannelCache)

SingleMessage = _reflection.GeneratedProtocolMessageType('SingleMessage', (_message.Message,), {
  'DESCRIPTOR' : _SINGLEMESSAGE,
  '__module__' : 'cyber.proto.record_pb2'
  # @@protoc_insertion_point(class_scope:apollo.cyber.proto.SingleMessage)
  })
_sym_db.RegisterMessage(SingleMessage)

Header = _reflection.GeneratedProtocolMessageType('Header', (_message.Message,), {
  'DESCRIPTOR' : _HEADER,
  '__module__' : 'cyber.proto.record_pb2'
  # @@protoc_insertion_point(class_scope:apollo.cyber.proto.Header)
  })
_sym_db.RegisterMessage(Header)

Channel = _reflection.GeneratedProtocolMessageType('Channel', (_message.Message,), {
  'DESCRIPTOR' : _CHANNEL,
  '__module__' : 'cyber.proto.record_pb2'
  # @@protoc_insertion_point(class_scope:apollo.cyber.proto.Channel)
  })
_sym_db.RegisterMessage(Channel)

ChunkHeader = _reflection.GeneratedProtocolMessageType('ChunkHeader', (_message.Message,), {
  'DESCRIPTOR' : _CHUNKHEADER,
  '__module__' : 'cyber.proto.record_pb2'
  # @@protoc_insertion_point(class_scope:apollo.cyber.proto.ChunkHeader)
  })
_sym_db.RegisterMessage(ChunkHeader)

ChunkBody = _reflection.GeneratedProtocolMessageType('ChunkBody', (_message.Message,), {
  'DESCRIPTOR' : _CHUNKBODY,
  '__module__' : 'cyber.proto.record_pb2'
  # @@protoc_insertion_point(class_scope:apollo.cyber.proto.ChunkBody)
  })
_sym_db.RegisterMessage(ChunkBody)

Index = _reflection.GeneratedProtocolMessageType('Index', (_message.Message,), {
  'DESCRIPTOR' : _INDEX,
  '__module__' : 'cyber.proto.record_pb2'
  # @@protoc_insertion_point(class_scope:apollo.cyber.proto.Index)
  })
_sym_db.RegisterMessage(Index)


# @@protoc_insertion_point(module_scope)



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="cyber_record",
    version="0.1.1",
    author="daohu527",
    author_email="daohu527@gmail.com",
    description="Cyber record offline parse tool",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/daohu527/cyber_record",
    project_urls={
        "Bug Tracker": "https://github.com/daohu527/cyber_record/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    package_dir={"": "."},
    packages=setuptools.find_packages(where="."),
    install_requires=[
        'protobuf>=3.20.0',
    ],
    python_requires=">=3.6",
)



#!/usr/bin/env python

# Copyright 2022 daohu527 <daohu527@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import time

from modules.map.proto import map_pb2
from cyber_record.record import Record
from record_msg.builder import ImageBuilder, PointCloudBuilder


def write_message():
  pb_map = map_pb2.Map()
  pb_map.header.version = 'hello'.encode()

  write_file_name = "example_w.record.00000"
  with Record(write_file_name, mode='w') as record:
    record.write('/apollo/map', pb_map, int(time.time() * 1e9))


def read_write_message():
  read_file_name = "example.record.00000"
  r_record = Record(read_file_name)

  write_file_name = "example_w.record.00001"
  with Record(write_file_name, mode='w') as w_record:
    for topic, message, t in r_record.read_messages_fallback():
      print("{}, {}, {}".format(topic, type(message), t))
      w_record.write(topic, message, t)


def write_image():
  image_builder = ImageBuilder()
  write_file_name = "example_w.record.00002"
  with Record(write_file_name, mode='w') as record:
    img_path = 'test.jpg'
    pb_image = image_builder.build(img_path, encoding='rgb8')
    record.write('/apollo/sensor/camera/front_6mm/image',
                 pb_image,
                 int(time.time() * 1e9))


def write_point_cloud():
  point_cloud_builder = PointCloudBuilder()
  write_file_name = "example_w.record.00003"
  with Record(write_file_name, mode='w') as record:
    pcd_path = 'test.pcd'
    pb_point_cloud = point_cloud_builder.build(pcd_path)
    record.write('/apollo/sensor/lidar32/compensator/PointCloud2',
                 pb_point_cloud,
                 int(time.time() * 1e9))


if __name__ == "__main__":
  write_message()
  read_write_message()
  write_image()
  write_point_cloud()
